/* $Workfile: T2p5x4Hdlc.c $       $Revision: 118 $      $Modtime: 5/05/08 1:57p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4HDLC.C

    A simple example console application demonstrating the use of OTX Driver.

    The applications works with the following board combinations:

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Gimle-16-PCI-Plus / Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-CST,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-EX,
      Gimle-16-PCI-Plus / Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-55x4-ASM,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Express + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Express + Vidar-5x4-ASM-EX,
      Thor-2-PCI-Express + Vidar-55x4-ASM,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or Sleipnir-1-PCI,

    or Hugin-2-PC104

    or Thor-2-ExpressCard.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures all available E1T1 devices on the board. The E1T1 spans should be
    looped with an external loop cable (e.g. Li0<->Li1, Li2<->Li3, etc).
    
    The  controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    The HDLC Sender/Receiver is hosted by the DSP device. Currently it is hardcoded
    to use only one HDLC channel per Li, but it could fairly easily be expanded
    to use more than one channel per Li (see coHdlcPerLi).

    Designed for 4, 2, and 1 DSPs (4 for Vidar-5x4-ASM, 2 for Thor-2-PCMCIA, 
    1 for Sleipnir-1-PCI, and 2 for Hugin-2-PC104).

    SS7 FISU filtering is supported (see the command line switches).

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
//#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_MAX_MULTIPLE_FRAME_READ 128

#define DEMO_SS7_SIM_FRAMES 1000

#define DEMO_DEFAULT_LOG_FILE "T2p5x4Hdlc.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL                bExitOnError;
  char                  *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  char                   cwd[OTX_MAX_PATH];
  int                    coDsp;
  int                    coLi;
  int                    coHdlcPerLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bSs7FisuFiltering;
  OTX_BOOL               bHdlcBitInversion;
  int                    boardNo;
  OTX_BOOL               bUseDspDebugger;
  OTX_BOOL               bSuperChannel;
  OTX_BOOL               b56kbpsChannel;
  int                    nTimeSlots;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxDspSpmE             eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OTX_BOOL               bUseSenders;
  OTX_BOOL               bUseReceivers;
  OtxBrdClockSourceE  eClkSource;
  int                    nSingleTs;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
} DemoEnvT;

void *DemoMsuTransmitFunction(
  void *lpParameter   // thread data
);

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoLogf2(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_INT32   g_cbHdlcFrame0 = 10;
OTX_INT32   g_cbHdlcFrame1 = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_BOOL    g_bHdlcCheckRxFrames = OTX_TRUE;
OTX_UINT32  g_coHdlcRxBytes[32];
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];
OTX_UINT32  g_coHdlcFramesSentRollOverCount[32];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 0;
OtxHdlcSS7FisuS g_tFisu0;
OtxHdlcSS7FisuS g_tFisu1;
OtxHdlcSS7ShortLssuS g_tLssu;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs

OTX_INT32   g_nSs7SimulationErrors[32] = { 0 };
OTX_BOOL    g_bSs7SimulationActive[32] = { 0 };
OTX_INT32   g_nSs7SimulationFramesReceived[32] = { 0 };
OTX_UINT8   g_nPrevBsn[32] = {0x7F};

OTX_UINT32  g_BsnErrors = 0;
OTX_BOOL    g_BsnCheck = OTX_FALSE;

OsaMutexS g_demoLogfCritSec;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[4]; // Only used for Thor-8-PCI, Thor-8-PCI-Plus and Gimle-16-PCI-Plus
  OTX_HANDLE hVidarAsmController0;  // Not used for Thor-2-PCMCIA
  OTX_HANDLE hLi[16];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
} g_sqDspHdlcSender[32];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
} g_sqDspHdlcReceiver[32];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


#if defined(_WIN32_WCE)
#define CLOCKS_PER_SEC 1
typedef int clock_t;
int clock() { return 0; }
#endif 
/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; 

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    break;
  case OTX_DEVICE_HUGIN_2_PC104:
    DemoLogf("Opening Hugin-2-PC104 board\n");
    break;
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    DemoLogf("Opening Thor-1-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This 
    // demo however needs the card to support the transmit data so the 
    // card will use 2 receivers and the 2 transmitters.
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n");

    // we'll source the Falc from the DSP device, and source the DSP from the FALC
    //  FALC <-- DSP
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_FALC_SOURCED, 
                                           OTX_BRD_FALC_SOURCED_DSP); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);
    //  DSP <-- FALC
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_DSP_SOURCED, 
                                           OTX_BRD_DSP_SOURCED_FALC); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);

    DemoLogf("  Set FPGA highway path DSP->Falc and Falc->DSP\n");

  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      // On the Gimle-16-PCI-Plus board the E1T1 devices are hosted by 4 quad E1T1 devices. Open those four devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {
      // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a Dual E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_DUAL_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = 0;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_DUAL_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Get handles to the physcial DSP device
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      char szDevString[32];
      if (i == 0) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
        result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
        DemoLogf("DSP#%d: %s\n", i, szDevString);
      } else {
        g_hPhysDevs.hDsp[i] = g_hPhysDevs.hDsp[0]; // Just to make a demo program generic for 4 DSPs. This card really only has one DSP.
      }
    } 
    break;
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_HUGIN_2_PC104:
    // Get handles to the physcial DSP devices
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS: 
  case OTX_DEVICE_THOR_2_PCI_EXPRESS: 
  case OTX_DEVICE_THOR_4_PCI_PLUS: 
  case OTX_DEVICE_THOR_8_PCI: 
  case OTX_DEVICE_THOR_8_PCI_PLUS: 
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_THOR_1_PCI_PLUS: 
    {
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
          DemoLogf("Opening Vidar-5x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
          DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:
          DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
          DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          bOpenAsm = OTX_TRUE;

          // Verify that the system supports the Vidar-55x4-ASM board
          {
            char szCpldBuildNumbers[256];
            char szBrdBuildNumber[16];
            OTX_INT32 nBoardBuildNo;

            if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
              
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
                DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
                DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
              
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 11) {
                DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }
              
            } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
                DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
                DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
              
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 5) {
                DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }
              
            }
            
          }


          break;
        }
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
        }
      }
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
/*
  if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {

    // we'll source the Falc from the DSP device, and source the DSP device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_DSP); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DSP_SOURCED, OTX_BRD_DSP_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }
*/

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);

    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD ||
        g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      // There is only one DSP on this card. 
      // No need to load the program more than once.
      i = g_demoEnv.coDsp; 
    }
  }


	// Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM1\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2EX:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLC:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLCEX:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLCEX\n", i);
      break;
    case OTX_DSP_SPM_OTXCST3:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXCST3\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLC_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC_INTMEM\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2_INTMEM\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2EX_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX_INTMEM\n", i);
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
#if 1
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#else
    DemoLogf("Loading from file\n");
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxSpm2.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxSpm2c5410a.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxHdlc5410a.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxHdlcEx5416.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#endif

    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD ||
        g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      // There is only one DSP on this card. 
      // No need to load the program more than once.
      i = g_demoEnv.coDsp; 
    }
  }


  if (g_demoEnv.bUseDspDebugger) {
    DemoLogf("Please load and run DSP program %s to DSP#%d-%d in your DSP Debugger\n", "OtxHdlc.out", 0, g_demoEnv.coDsp-1);
    DemoLogf("Press any key when all the DSP programs are running...\n");
    OsaConsoleGetCh();
    
    // At this point we need to tell the host to restart the programs 
    // (to issue the initial IO commands to the DSP again)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      DemoLogf("Restarting program on DSP#%d\n", i);
      result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
  }


}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Initialize the T1/E1 Line Interface Devices

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Set line termination mode
    {
      
      DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
      switch (g_demoEnv.eTermMode) {
      case OTX_T1E1_TERM_120_OHM:
        DemoLogf("OTX_T1E1_TERM_120_OHM\n");
        break;
      case OTX_T1E1_TERM_100_OHM:
        DemoLogf("OTX_T1E1_TERM_100_OHM\n");
        break;
      case OTX_T1E1_TERM_75_OHM:
        DemoLogf("OTX_T1E1_TERM_75_OHM\n");
        break;
      case OTX_T1E1_TERM_HIGHZ:
        DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
        break;
      default:
        DemoLogf("Invalid Termination mode.\n");
        break;
      }
      
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      
      if (g_demoEnv.eTermMode == OTX_T1E1_TERM_HIGHZ) {
        DemoLogf("Li[%d]: Tristating Transmit pins in High-Z Termination mode\n", i);
        
        // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
        result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      } else {

        if(g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCI_PLUS) {
          // Drive the transmit pins of the Line interface in all other termination modes
          result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        }
      }
      
    }
    
    
    // Set receive threshold
    DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    { // Read back the threshold
      OTX_UINT32 nRxThreshold;
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
    }
    
    // Set Monitor (-20dB) mode
    DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    
    // Turning front end Amplifier ON/OFF      
    DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    


  }

}

/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; 

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board has Quad E1T1 devices then enable them
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    {      
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {      
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {
      // Enable Quad E1T1 devices
      nQuadE1T1 = 0;
      DemoLogf("Enable OTX_DEVICE_DUAL_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;

  default:
    // Nothing
    break;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
#if 0
  OTX_UINT32 nIrpQueueLimit;
#endif


  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
    if (g_demoEnv.bUseSenders) {
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);
    }
    if (g_demoEnv.bUseReceivers) {
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], result);
    }

#if 0
    nIrpQueueLimit = 1;
    DemoLogf("HdlcSender[%d]: Setting nIrpQueueLimit = %d\n", i, nIrpQueueLimit);
    result = OtxDrvSetAttributeValueUint32(g_sqDspHdlcSender[i].hHandle, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, nIrpQueueLimit); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
#endif
  }
  
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqLiTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];
  OTX_INT32 nLi;
  OTX_INT32 nTsLi;
  OTX_INT32 nTsDsp;
  OTX_UINT32 coTsPerDspHwy = g_nDspHighwaySpeed / 64;

  if (g_demoEnv.bSuperChannel) {
    // x*64kbps HDLC capacity (TS1-TS31 E1 only)
    OTX_INT32 j;
    OTX_INT32 noOfTimeSlots = g_demoEnv.nTimeSlots;
    i = 0;

    DemoLogf("Configuring one %dkbps HDLC channel per Li\n", 64*g_demoEnv.nTimeSlots);

    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++, i++) {
      
      assert(g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_E1);
      
      memset(sqLiTimeSlotMask, 0x00, 32 * sizeof(OTX_UINT8));
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));

      for (j = 0; j < noOfTimeSlots; j++) {
        // Timeslot on the E1 span
        nTsLi = 1+j;

        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
          // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
          nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
        } else
        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
          // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
          // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
          // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
          nTsDsp = (nTsLi * 4) + nLi ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
        } else {
          nTsDsp = nTsLi;
        }
        sqLiTimeSlotMask[nTsLi] = 0xFF;
        sqDspTimeSlotMask[nTsDsp] = 0xFF;

      }

      if (g_demoEnv.bUseSenders) {
        result = OtxDrvCreatePipe(noOfTimeSlots*64, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
        
        // Timeslot on the DSP
        //memset(sqLiTimeSlotMask, 0xFF, 32 * sizeof(OTX_UINT8));
        //sqLiTimeSlotMask[0] = 0x00;
        result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
        result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
        
        // Make the following connections in transmit direction:
        // HDLC Sender   -----> Li
        result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      }      
      
      if (g_demoEnv.bUseReceivers) {
        result = OtxDrvCreatePipe(noOfTimeSlots*64, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        
        result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[(i+2)%g_demoEnv.coDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        
        // Create the following Connections in receive direction:
        //  Li  ------> HDLC Receiver
        result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
        
      }
    }
  } else { 
    // 64kbps or 56kbps HDLC capacity
    OTX_INT32 j;
    OTX_INT32 nTsLi;
    OTX_INT32 nCapacity;
    OTX_UINT8 nMaskValue;
    
    if (g_demoEnv.b56kbpsChannel) {
      DemoLogf("Configuring one 56kbps HDLC channel per Li\n");
      nCapacity = 56;
    } else {
      DemoLogf("Configuring one 64kbps HDLC channel per Li\n");
      nCapacity = 64;
    }

    i = 0;
    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
      if (g_demoEnv.nSingleTs == -1) {
        // Timeslot not specified on the command line, so pick a default timeslot depending on the T1/E1 mode
        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nTsLi = 24;
        } else {
          // E1 mode
          nTsLi = 16;
        }
      } else {
        nTsLi = g_demoEnv.nSingleTs;
      }
      
      for (j = 0; j < g_demoEnv.coHdlcPerLi; j++, i++, nTsLi--) {
        
        // Connect pipes to the E1/T1 device
        memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));

        if (g_demoEnv.b56kbpsChannel) {
          nMaskValue = 0xFE; 
        } else {
          nMaskValue = 0xFF; 
        }

        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
          // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
          nTsDsp = nTsLi * 4 + nLi * 2 ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
        } else 
        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
          // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
          // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
          // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
          nTsDsp = (nTsLi * 4) + nLi;  // See comment above regarding the timeslot mapping for the DSP highway on this board
        } else {
          nTsDsp = nTsLi;
        }

        sqLiTimeSlotMask[nTsLi] = nMaskValue; // Timeslot on the T1/E1 span
        sqDspTimeSlotMask[nTsDsp] = nMaskValue;
   
        if (g_demoEnv.bUseReceivers) {
          result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
          
          g_sqDspHdlcReceiver[i].nLi = nLi;
          result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
          
          g_sqDspHdlcReceiver[i].nDsp = (i+2)%g_demoEnv.coDsp;
          result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[g_sqDspHdlcReceiver[i].nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
          
          // Create the following Connections in receive direction:
          //  Li  ------> HDLC Receiver
          result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
          
          DemoLogf("Hdlc Receiver[%d] connected (Dsp#%d, Li#%d, TSLi%d, TSDsp%d) \n", i, g_sqDspHdlcReceiver[i].nDsp, g_sqDspHdlcReceiver[i].nLi, nTsLi, nTsDsp);
        }
        
        
        if (g_demoEnv.bUseSenders) {
          result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
          
          g_sqDspHdlcSender[i].nLi = nLi;
          result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
          
          g_sqDspHdlcSender[i].nDsp = i%g_demoEnv.coDsp; 
          result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
          
          // Make the following connections in transmit direction:
          // HDLC Sender   -----> Li
          result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
            if(g_sqDspHdlcSender[i].nLi < 2) {
              DemoLogf("Hdlc Sender[%d]   connected (Dsp#%d, Li#%d, TSLi%d, TSDsp%d) \n", i, g_sqDspHdlcSender[i].nDsp, g_sqDspHdlcSender[i].nLi, nTsLi, nTsDsp);
            }
            else {
              DemoLogf("Hdlc Sender[%d] Board Type %d does not support transmit on Li#%d \n", i, g_demoEnv.nBoardType, g_sqDspHdlcSender[i].nLi);
            }
          }
          else {
            DemoLogf("Hdlc Sender[%d]   connected (Dsp#%d, Li#%d, TS%d) \n", i, g_sqDspHdlcSender[i].nDsp, g_sqDspHdlcSender[i].nLi, nTsLi);
          }
        } // if (g_demoEnv.bUseSenders)       
      } // for  j < g_demoEnv.coHdlcPerLi
    } // for nLi < g_demoEnv.coLi
  } // g_demoEnv.bSuperChannel
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
    if (g_demoEnv.bUseSenders) {    
      
#if 0
/*
      // Reduce the IRP queue to 1 for the HDLC Sender 
      // (which is the number of HDLC frames that the driver will 
      // buffer before returning OTX_E_BUSY)
      // The default IRP queue size is 10 frames.
      result = OtxDrvSetAttributeValueUint32(
        g_sqDspHdlcSender[i].hHandle, 
        OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, 
        1); 
      DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
*/
#endif
      
      
      // Initialize the Hdlc Sender 
      result = OtxHdlcSenderInit(
        g_sqDspHdlcSender[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS/*OTX_HDLC_FFT_ALL_ONES*/,        // Fill character(s) to be send between HDLC frames
        1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
        OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
        ); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);          
    } // if g_demoEnv.bUseSenders
        
    if (g_demoEnv.bUseReceivers) {
      
      // Initialize the Hdlc Receiver 
      result = OtxHdlcReceiverInit(
        g_sqDspHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS/*OTX_HDLC_FFT_ALL_ONES*/,        // Fill character(s) to be send between HDLC frames
        1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
        OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
        OTX_TRUE                  // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
        ); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);          
      
      if (g_demoEnv.bSs7FisuFiltering) {
        //if (i == 1) continue;
        result = OtxHdlcSS7SetFilter(g_sqDspHdlcReceiver[i].hHandle, OTX_HDLC_SS7_FILTER_FISU | OTX_HDLC_SS7_FILTER_SHORT_LSSU | OTX_HDLC_SS7_FILTER_LONG_LSSU); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
      }
    } // if g_demoEnv.bUseReceivers
  } // for i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi
}



/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  if (g_demoEnv.bUseSenders) {
    for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
      //DemoLogf("** i=%d Enabling hSpm1Sender[%d] \n", i); OsaConsoleOsaConsoleGetCh();
      result = OtxDrvEnable(g_sqDspHdlcSender[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
    }
  }

  if (g_demoEnv.bUseReceivers) {
    for (i = 0; i < g_demoEnv.coLi*g_demoEnv.coHdlcPerLi; i++) {
      //DemoLogf("** i=%d Enabling hSpm1Receiver[%d] \n", i); OsaConsoleOsaConsoleGetCh();
      result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
    }
  }
  
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
#if 1
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
#else
  pE1CfgData->eAisDetection              = OTX_E1_AIS_G775;
#endif
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000;
}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;

  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    //DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

#if 0
  if (g_bHdlcVerbose) {
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
  }

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }
#endif

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
       DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);

    //	DemoLogf("DSP[%d] Event: %s (m_nParam=%d, m_nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

#define DEMO_HDLC_RX_BUFFER_SIZE 2048
#if 1 // Single Frame Read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[DEMO_HDLC_RX_BUFFER_SIZE];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OtxHdlcFrameEndCodeE nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      OTX_RESULT result;
      OtxHdlcSS7FisuS *pFisu;
      OTX_INT32 nDev; 
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
          g_coHdlcRxBytes[pEventData->m_nSourceId] += nFrameLen;
        } else {
          DemoLogf("HDLC Frame Received with an invalid m_nSourceId (%d)\n", pEventData->m_nSourceId);
        }

        nDev = pEventData->m_nSourceId;
        if (g_bSs7SimulationActive[nDev]) {
          if (nFrameEndCode == 0) {
            OTX_BOOL bErroredFrame = OTX_FALSE;

            pFisu = (OtxHdlcSS7FisuS *)&buf[0];

            if (((OTX_UINT8)pFisu->bsn != g_nPrevBsn[nDev]) && ((OTX_UINT8)pFisu->bsn != ((g_nPrevBsn[nDev]+1) % 128))) {
              bErroredFrame = OTX_TRUE;
              g_nSs7SimulationErrors[nDev]++;
              DemoLogf2("=============v Bsn Error (%d)\n", g_nSs7SimulationErrors[nDev]);
            }

            if (nFrameLen == 3) {
              // FISU
              // No further check
            } else {
              // MSU
              
              if (nFrameLen != (OTX_UINT32)(((g_nSs7SimulationFramesReceived[nDev]%10)*10) + 10)) {
                // Data length check
                bErroredFrame = OTX_TRUE;
                g_nSs7SimulationErrors[nDev]++;
                DemoLogf2("=============v Length Error (%d). Expected %d. Got %d.\n", g_nSs7SimulationErrors[nDev], (((g_nSs7SimulationFramesReceived[nDev]%10)*10) + 10), nFrameLen);
              } else {
                // Payload check
                for (i = 3; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
                  if (buf[i] != (OTX_UINT8)i) {
                    bErroredFrame = OTX_TRUE;
                    g_nSs7SimulationErrors[nDev]++;
                    DemoLogf2("=============v Payload Error (%d). Expected %d. Got %d.\n", g_nSs7SimulationErrors[nDev], i, buf[i]);
                  }
                }
              }


              g_nSs7SimulationFramesReceived[nDev]++;

            }

            //if (bErroredFrame) {
              DemoLogf2("%d:Len=%.3d  ", pEventData->m_nSourceId, nFrameLen);
              for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
                DemoLogf2("%.2X ", buf[i]);
              }
              DemoLogf2("\n");
            //}

            g_nPrevBsn[nDev] = pFisu->bsn;
            

          } else {
            g_nSs7SimulationErrors[nDev]++;
            DemoLogf2("%d:Len=%.3d  Bad Frame\n", pEventData->m_nSourceId, nFrameLen);
          }


          if (g_nSs7SimulationFramesReceived[nDev] == DEMO_SS7_SIM_FRAMES) {
            OTX_INT32 nLi = g_sqDspHdlcSender[pEventData->m_nSourceId].nLi;
            DemoLogf("%d Received %d frames. Receive SS7 Done. Errors=%d\n", pEventData->m_nSourceId, DEMO_SS7_SIM_FRAMES, g_nSs7SimulationErrors[nDev]);
            g_bSs7SimulationActive[nDev] = OTX_FALSE;
            DemoLogf("%d Stopping FISU Sending for HDLC Sender %d\n", pEventData->m_nSourceId, pEventData->m_nSourceId);
            result = OtxHdlcStopRepeatData(g_sqDspHdlcSender[pEventData->m_nSourceId].hHandle); 
            result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOOPS_OFF); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
          } 
#if 0
        /*
        } else if (pEventData->m_nSourceId == 1) {
          nFrameCount++;

          if (nFrameEndCode == 0) {
            static OTX_BOOL getReady = OTX_FALSE; 
            
            if (g_BsnCheck) {
              if (nFrameLen == 5) {
                g_BsnCheck = OTX_FALSE;
                DemoLogf2("Turning checking off\n");
              } else if (nFrameLen == 4) {
                getReady = OTX_TRUE;
              } else if (getReady) {
                getReady = OTX_FALSE;
                if (nFrameCount > 1) {
                  if (nPrevBsn == buf[0]) {
                    g_BsnErrors++;
                    DemoLogf2("=============> Bsn Error (%d)\n", g_BsnErrors);
                  }
                }
                nPrevBsn = buf[0];
              }
            }

            DemoLogf2("%d:Len=%.3d  ", pEventData->m_nSourceId, nFrameLen);
            for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
              DemoLogf2("%.2X ", buf[i]);
            }
            DemoLogf2("\n");
          } else {
            DemoLogf2("%d:Len=%.3d  Bad Frame\n", pEventData->m_nSourceId, nFrameLen);
          }
          */
#endif
        } else {
        
          if (g_bHdlcVerbose) {
            //DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            DemoLogf("HdlcReceiver[%d]  \n", pEventData->m_nSourceId);
            
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            //DemoLogf("[%d] Len=%.3d, Fe=0x%.2X   ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode);
            for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
          } else {
            if (nFrameEndCode) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                result = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, result);
              }
              
            }
          }
#if 0
          /*
          {
            static OTX_INT32 nHdlcDev = 0;
            OtxHdlcSS7FisuS *pFisuBuf = (OtxHdlcSS7FisuS *) buf;

            g_tFisu0.bsn = pFisuBuf->fsn;
            g_cbHdlcFrame0 = 3;
            OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &nHdlcDev);
          }
          */
#endif

        }
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      }
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
    } else {
      // Unexpected event from this device
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#else // Multiple frame read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId] += coFrames;
        }

        if (g_bHdlcVerbose) {
          DemoLogf("%d HDLC Frame(s) Received, Sequence Number = %d\n", coFrames, pEventData->m_nParam);
        }
        
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

					pBuf = pFrame->sqData;

          if (g_bHdlcVerbose) {
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->nFrameSeq, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf("%.2X ", pBuf[i]);
            }
            DemoLogf("\n");
          } else {

            if (pEventData->m_nSourceId < 32) {
              g_coHdlcRxBytes[pEventData->m_nSourceId] += pFrame->cbOctets;
            }

            if (pFrame->nFrameEndCode) {
              
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }
              
              
            } else if (g_bHdlcCheckRxFrames) {
              if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
                
              } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              }
            }
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);           }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#endif


  case OTX_LDEVICE_HDLC_SENDER:


    if(pEventData->m_nCode == OTX_HDLC_SENDER_EC_FRAME_COUNT_ZERO){
      // The counter will rollover ever 65536 * 125uSec or 8.192 seconds.
      g_coHdlcFramesSentRollOverCount[pEventData->m_nSourceId]++;
      /* Event is generated a lot, printing every time will fill the display.
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      DemoLogf(" OTX_HDLC_SENDER_EC_FRAME_COUNT_ZERO time slot %d, Rollover %d\n", 
               pEventData->m_nParam,
               g_coHdlcFramesSentRollOverCount[pEventData->m_nSourceId]);
      */
    }
    else
    if(pEventData->m_nCode == OTX_HDLC_SENDER_EC_FRAME_SENT){
      OTX_UINT16 *pData;

      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      // Dispaly the 16 bit frame sent count value. This value added to the 
      // g_coHdlcFramesSentRollOverCount multiplied by 65536 is the total  
      // number of 125 microSec frames that have been sent.
      pData = (OTX_UINT16 *) &pEventData->m_nParam;
      DemoLogf("  OTX_HDLC_SENDER_EC_FRAME_SENT frame count %d, Frame Slot %d\n", 
               pData[0], pData[1]);
      DemoLogf("g_coHdlcFramesSentRollOverCount count %d \n", 
               g_coHdlcFramesSentRollOverCount[pEventData->m_nSourceId]);
    }
    else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    }

    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}



/*******************************************************************************

  @func

  Transmits one or more HDLC frames

  @comm

*******************************************************************************/
void *DemoHdlcTransmitFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OTX_UINT32 nDev;

  nDev = *((OTX_INT32 *) lpParameter);
  
#if 1 
  for (i = 0; i < g_cbHdlcFrame0; i++) {
    g_sqHdlcFrame[i] = (OTX_UINT8) i;
  }
#else // Testing zero-bit insertion of the last bit of the frame 
  g_cbHdlcFrame0 = 3;
  g_sqHdlcFrame[0] = 0xEC;
  g_sqHdlcFrame[1] = 0x49;
  g_sqHdlcFrame[2] = 0x00;
#endif
  
  DemoLogf("Sending %d HDLC frame(s) from HDLC Sender #%d (len=%d)\n", g_coHdlcSenderFrames, nDev, g_cbHdlcFrame0);
    
  // Send HDLC Frame(s)
  startTime = clock();
  for (i = 0; i < g_coHdlcSenderFrames; i++) {
    result = OtxHdlcSendData(g_sqDspHdlcSender[nDev].hHandle, g_sqHdlcFrame, g_cbHdlcFrame0, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      i--;
      busyCount++;
    } else {
      DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);
      OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
      //nHdlcTaskRef++;
    }
  }
  stopTime = clock();
  
  printf("[%d]Done (BusyCount=%d, time=%2.1f seconds).\n", nDev, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);

  return 0;

}




/*******************************************************************************

  @func

  Transmits one or more MSU frames

  @comm

*******************************************************************************/
void *DemoMsuTransmitFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS *pFisu;
  OTX_UINT8 sqHdlcFrame[300];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;

  nDev = *((OTX_INT32 *) lpParameter);

  
  // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
  if (nDev == 0) {
    pFisu = &g_tFisu0;
    cbHdlcFrame = g_cbHdlcFrame0;
  } else {
    pFisu = &g_tFisu1;
    cbHdlcFrame = g_cbHdlcFrame1;
  }

  memcpy(&msu, pFisu, OTX_HDLC_SS7_CB_FISU);  // Build the first 3 bytes of the MSU
  msu.li = (OTX_UINT8) cbHdlcFrame - OTX_HDLC_SS7_CB_FISU;  // Set the Length Indicator
  memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
  for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {     // Add the payload the HDLC frame buffer
    sqHdlcFrame[i] = (OTX_UINT8) i;
  }
  
  DemoLogf("Sending %d MSU frame(s) (len=%d) from HDLC Sender #%d  (Dsp#%d, Li#%d)\n", g_coHdlcSenderFrames, cbHdlcFrame, nDev, g_sqDspHdlcSender[nDev].nDsp, g_sqDspHdlcSender[nDev].nLi);
  
  // Send MSU Frame(s)
  startTime = clock();
  for (i = 0; i < g_coHdlcSenderFrames; i++) {
    result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, pFisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      i--;
      busyCount++;
    } else {
      DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);
      OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
    }
  }
  stopTime = clock();
  
  DemoLogf("[%d]Done (BusyCount=%d, time=%2.1f seconds).\n", nDev, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);

  return 0;
}



/*******************************************************************************

  @func

  Transmits one or more MSU frames

  @comm

*******************************************************************************/
void *DemoSs7SimulationFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS fisu;
  OTX_UINT8 sqHdlcFrame[200];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;
  OTX_INT32 coHdlcSenderFrames = DEMO_SS7_SIM_FRAMES;
  OTX_INT32 coMsusSent = 0;
  OTX_INT32 coFisusSent = 0;

  nDev = *((OTX_INT32 *) lpParameter);

  fisu.bib = 1;
  fisu.fib = 1;
  fisu.fsn = 0x7F;
  fisu.bsn = 0x7F;
  fisu.li = 0;
  fisu.reserved = 0;
  
  g_bSs7SimulationActive[nDev] = OTX_TRUE;

  memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
  cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
  
  DemoLogf("Sending a mix of MSU and FISU frames from HDLC Sender #%d  (Dsp#%d, Li#%d)\n", nDev, g_sqDspHdlcSender[nDev].nDsp, g_sqDspHdlcSender[nDev].nLi);
  
  // Send MSU Frame(s)
  startTime = clock();
  while (coMsusSent < coHdlcSenderFrames) {

    result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, &fisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      busyCount++;
      OsaThreadSleep(1);
    } else {
      DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);

      // Step BSN for every new frame package (MSU+FISU or FISU+FISU) to send
      fisu.bsn++;

      if (coFisusSent == 0) { 
        OsaThreadSleep(15);
      } else {
        OsaThreadSleep(rand() % 5);
      }

      if (cbHdlcFrame == 3) {
        coFisusSent++;
      } else {
        DemoLogf2("%d: Sent MSU# %d\n", nDev, coMsusSent);
        coMsusSent++;
      }

      if ((rand() % 100) < 50) {
        // Prepare an MSU
        cbHdlcFrame = ((coMsusSent%10)*10) + 10;
        memcpy(&msu, &fisu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        msu.li = (OTX_UINT8)cbHdlcFrame - OTX_HDLC_SS7_CB_FISU;  // Set the Length Indicator
        memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {     // Add the payload the HDLC frame buffer
          sqHdlcFrame[i] = (OTX_UINT8) i;
        }


      } else {
        // Prepare a FISU
        memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
        cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
      }

    }
  }
  stopTime = clock();
  
  DemoLogf("[%d]Done (BusyCount=%d, time=%2.1f seconds) FISUs=%d MSUs=%d.\n", nDev, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC, coFisusSent, coMsusSent);

  return 0;
}






/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;
  envP->coHdlcPerLi       = 1; // for now
  envP->coLi              = 2;
  envP->bUseDspDebugger   = OTX_FALSE;
  envP->coDsp             = 4;
  envP->bHdlcBitInversion = OTX_FALSE;
  envP->bSuperChannel     = OTX_FALSE;
  envP->b56kbpsChannel    = OTX_FALSE;
  envP->nTimeSlots        = 11;
  envP->nSingleTs         = -1;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLC;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->bUseSenders       = OTX_TRUE;
  envP->bUseReceivers     = OTX_TRUE;
  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'I') {   // HDLC Bit Inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i]+1) == 'X') {   // Use Super-channel (higher than 64kbps capacity)
          envP->bSuperChannel = OTX_TRUE;
        } else if (*(argv[i]+1) == '5') {   // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // Number of timeslots in the superchannel
          envP->nTimeSlots = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'V') {   // Timeslot to connect the HDLC device to (for 64/56 kbps channels)
          envP->nSingleTs = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   // Line Interface 1
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == 'E') {   // External
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   // PCMCIA:Split clock, both Lis synchronized to its own line
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   // PCMCIA:Split clock, Li0 synchronized to its T1E1 span. Li1 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   // PCMCIA:Split clock, Li1 synchronized to its T1E1 span. Li0 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   // Li6 (if available)
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   // Li7 (if available)
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else if ( (*(argv[i]+2) == 'H') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == '1') {          // OtxSpm1.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM1;
          } else if (*(argv[i]+2) == '2') {   // OtxSpm2.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else if (*(argv[i]+2) == 'X') {   // OtxSpm2Ex.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
          } else if (*(argv[i]+2) == 'H') {   // OtxHdlc.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC;
          } else if (*(argv[i]+2) == 'E') {   // OtxHdlcEx.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLCEX;
          } else if (*(argv[i]+2) == '3') {   // OtxCst3.out
            envP->eProgramModule = OTX_DSP_SPM_OTXCST3;
          } else if (*(argv[i]+2) == 'G') {   // OtxHdlcEx_Intmem.out (Thor-2-ExpressCard)
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC_INTMEM;
          } else if (*(argv[i]+2) == 'N') {   // OtxSpm2_Intmem.out (Thor-2-ExpressCard)
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2_INTMEM;
          } else if (*(argv[i]+2) == 'M') {   // OtxSpm2Ex_Intmem.out (Thor-2-ExpressCard)
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX_INTMEM;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coLi = 2;
            envP->coDsp = 4;
            envP->coHdlcPerLi = 12; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi = 2;
            envP->coDsp = 2;
            //envP->coHdlcPerLi = 6; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
            envP->coHdlcPerLi = 1; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
            envP->coLi = 8;
            envP->coDsp = 4;
            envP->coHdlcPerLi = 3; // 6 Sender and 6 Receivers per DSP is the maximum number the OtxHdlc can handle
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; 
            envP->bUseSenders = OTX_FALSE;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coDsp = 1;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == 'H') {
            envP->nBoardType  = OTX_DEVICE_HUGIN_2_PC104;
            envP->coLi = 2;
            envP->coDsp = 2;
            envP->coHdlcPerLi = 1; // For now. This can be increased.
          } else if ( *(argv[i]+2) == '1') {
            envP->nBoardType  = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->coLi = 1;
            envP->coDsp = 1;
            envP->coHdlcPerLi = 1; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coLi = 2;
            envP->coDsp = 2; // Not really. this card only has 1 DSP.
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coLi  = 4;
            envP->coDsp = 2; // Not really. this card only has 1 DSP.
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else if ( *(argv[i]+2) == 'J') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF_J1; // J1_ESF
          } else if ( *(argv[i]+2) == '5') {
            envP->eT1FrameFormat = OTX_T1_FF_F4_J1; 
          } else if ( *(argv[i]+2) == 'F') {
            envP->eT1FrameFormat = OTX_T1_FF_F12_J1; // J1_SF
          } else if ( *(argv[i]+2) == '0') {
            envP->eT1FrameFormat = OTX_T1_FF_F72_J1; // J1_SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
          envP->bSs7FisuFiltering = OTX_TRUE;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

  if (g_demoEnv.bSuperChannel) {
    envP->coHdlcPerLi       = 1; // for higher capacity pipes we support only one HDLC channel per DSP
  }
  if (g_demoEnv.b56kbpsChannel) {
    envP->coHdlcPerLi       = 1;
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>   : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|T|E|M|4|8|P|G|S|H|C]  : Board Type used for the demo, where\n");
  DemoLogf("                      2 = Use Thor-2-PCI with Vidar-ASM\n");
  DemoLogf("                      T = Use Thor-2-PCI-Plus with Vidar-ASM Board\n");
  DemoLogf("                      E = Use Thor-2-PCI-Express with Vidar-ASM Board\n");
  DemoLogf("                      M = Use Thor-2-PCMCIA\n");
  DemoLogf("                      4 = Use Thor-4-PCI-Plus with Vidar-ASM Board\n");
  DemoLogf("                      8 = Use Thor-8-PCI with Vidar-ASM Board\n");
  DemoLogf("                      P = Use Thor-8-PCI-Plus with Vidar-ASM Board\n");
  DemoLogf("                      G = Use Gimle-16-PCI-Plus with Vidar-ASM Board\n");
  DemoLogf("                      S = Use Sleipnir-1-PCI (DSP version)\n");
  DemoLogf("                      H = Use Hugin-2-PC104\n");
  DemoLogf("                      C = Use Thor-2-ExpressCard Board\n");
  DemoLogf(" /L<fileName>  : Log file name\n");
  DemoLogf(" /S            : SS7 FISU filtering active\n");
  DemoLogf(" /ME           : E1 Li mode\n");
  DemoLogf(" /MT           : T1/J1 Li mode\n");
  DemoLogf(" /D            : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /I            : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  DemoLogf(" /5            : Use HDLC 56kbps channel.\n");
  DemoLogf(" /X            : Use HDLC Super-channels (higher than 64kbps capacity). Only for OTXHDLC.\n");
  DemoLogf(" /N<x>         : Number of 64kbps timeslots in the Super-channels (only used if /X is used)\n");
  DemoLogf(" /V<ts>        : E1/T1 Timeslot to connect the HDLC device to (for 64 or 56 kbps capacity)\n");
  DemoLogf(" /P[1|2|X|C|H|E|3|G|N|M] : DSP program to load, where\n");
  DemoLogf("                   1 = OtxSpm1\n");
  DemoLogf("                   2 = OtxSpm2\n");
  DemoLogf("                   X = OtxSpm2Ex\n");
  DemoLogf("                   H = OtxHdlc (default)\n");
  DemoLogf("                   E = OtxHdlcEx (for 5416 and 5510 DSPs)\n");
  DemoLogf("                   3 = OtxCst3 (for 54CST DSPs)\n");
  DemoLogf("                   G = OtxHdlc_IntMem (Thor-2-ExpressCard) \n");
  DemoLogf("                   N = OtxSpm2_IntMem (Thor-2-ExpressCard) \n");
  DemoLogf("                   M = OtxSpm2Ex_IntMem (Thor-2-ExpressCard) \n");
  DemoLogf(" /A            : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E            : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2            : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H            : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y            : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced (default)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /F[E|4|S|9|J|5|F|0] : T1/J1 Frame Format, where\n");
  DemoLogf("                   E = ESF (default)\n");
  DemoLogf("                   4 = F4\n");
  DemoLogf("                   S = SF [D3/D4] (F12)\n");
  DemoLogf("                   9 = SLC96 (F72)\n");
  DemoLogf("                   J = J1_ESF\n");
  DemoLogf("                   5 = J1_F4\n");
  DemoLogf("                   F = J1_SF\n");
  DemoLogf("                   0 = J1_SLC96\n");
  DemoLogf(" /J[A|B]       : T1/J1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]     : E1 Frame Format, where\n");
  DemoLogf("                   D = Doubleframe\n");
  DemoLogf("                   C = CRC4 Multiframe (default)\n");
  DemoLogf("                   G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]       : E1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   H = HDB3 (default)\n");
  DemoLogf(" /Z<str>       : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q            : Exit without prompting on errors\n");
  DemoLogf(" /K[x]         : Clock source, where\n");
  DemoLogf("                  x = I Internal clock source (default) or\n");
  DemoLogf("                  x = 0 through 7 indicating with Li to recover the clock from or\n");
  DemoLogf("                  x = H to recover from H.100 A-clock (ETCF mode)\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);

}

/*******************************************************************************

  @func

  prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf2(char *formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentHdlcDev = 0;
  OTX_RESULT result;

  DemoLogf("\nKey %c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis (Current HDLC device %d)\n", nCurrentHdlcDev);
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    if (g_demoEnv.bUseSenders) {
      
      DemoLogf(" a   Setting frame sent attribute for current HDLC Sender\n");
      DemoLogf(" t   Send an HDLC frame from current HDLC Sender\n");
      DemoLogf(" f   Start sending FISUs (continuosly) from current HDLC Sender\n");
      DemoLogf(" i   Start sending LSSUs (continuosly) from current HDLC Sender\n");
      DemoLogf(" m   Send an MSU from current HDLC Sender\n");
      DemoLogf(" F   Toggle on/off OTX_HDLC_SS7_FILTER_FISU_ALL filter current HDLC Sender\n");
      DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
      DemoLogf(" S   Set time to sleep (in ms) between transmitted HDLC Frames (%d)\n", g_nSleepTimeBetweenHdlcFrames);
      DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
    }
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" s   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
    DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
    DemoLogf(" p   Reset current HDLC Receiver\n");
    DemoLogf(" v   Toggle verbose logging of received HDLC frames\n");
    DemoLogf(" V   Print counters of received HDLC frames\n");
    DemoLogf(" w   Toggle payload checking of received frames\n");
    DemoLogf(" $   Query rx line frequency of the current clock source\n");
    break;

#if 1 // Only for debugging
  case 'x':
    {
      char szBuf[12];
      OTX_INT32 nDsp;                                                             
      
      DemoLogf("Dump profiling data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 3);
    }
    break;
  case 'j':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Dump debug data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
    }
    break;
  case 'J':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Dump debug data for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;
  case 'k':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 0;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;
  case 'K':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 1;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;

#endif
  case 'y':
    {
      OTX_UINT32 nAttribValue;
      DemoLogf("Fifo Diagnostics for HDLC Receiver device#%d:\n", nCurrentHdlcDev);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in frames)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in octets)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in frames)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in octets)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in frames): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in octets): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_LOST_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Number of lost frames    : %d\n", nAttribValue);
    }
    break;


  case 'w':
    {
      g_bHdlcCheckRxFrames = !g_bHdlcCheckRxFrames;
      DemoLogf("g_bHdlcCheckRxFrames is now %s.\n", g_bHdlcCheckRxFrames ? "TRUE": "FALSE");
    }
    break;

  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 16; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcRxBytes[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;

      static clock_t startTime = 0;
      double elapsedTime;
      static OTX_INT32 sqBytesReceived[16];
  
      for (i = 0; i < 16; i++) {
        if (g_coHdlcRxFrames[i]) {

          if (!startTime) {
            DemoLogf("[%.2d]  Tot bytes=%.10d Tot frames=%.6d (Bad=%.6d)\n", i, g_coHdlcRxBytes[i], g_coHdlcRxFrames[i], g_coHdlcBadRxFrames[i]);
          } else {
            elapsedTime = (double)(clock() - startTime) / CLOCKS_PER_SEC;

            DemoLogf("[%.2d]  Tot bytes=%.10d  %6.1f bytes/sec  Tot frames=%.6d (Bad=%.6d)\n", i, g_coHdlcRxBytes[i], (double)(g_coHdlcRxBytes[i]-sqBytesReceived[i])/elapsedTime, g_coHdlcRxFrames[i], g_coHdlcBadRxFrames[i]);
          }
        }
      }

      // Need a critical section here
      memcpy(sqBytesReceived, g_coHdlcRxBytes, sizeof(OTX_INT32) * 16);
      startTime = clock();

    }
    break;

  case 's':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for HdlcReceiver #%d \n", nCurrentHdlcDev);

      result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);

      if (OTX_SUCCESS(result)) {
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

  case 'g':
    {
      OTX_UINT32  nFisuCount0;   // Number of FISUs received
      OTX_UINT32  nLssuCount0;   // Number of LSSUs received
      OTX_UINT32  nMsuCount0;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu0;     // The last received FISU on the pipe.
      OTX_UINT32  nFisuCount1;   // Number of FISUs received
      OTX_UINT32  nLssuCount1;   // Number of LSSUs received
      OTX_UINT32  nMsuCount1;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu1;     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for HdlcReceiver #%d and #%d\n", 0, 1);

      result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[0].hHandle, &nFisuCount0, &nLssuCount0, &nMsuCount0, &fisu0); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[1].hHandle, &nFisuCount1, &nLssuCount1, &nMsuCount1, &fisu1); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);

      if (OTX_SUCCESS(result)) {
        DemoLogf("FISUs received: %.10lu    %.10lu  \n", nFisuCount0, nFisuCount1);
        DemoLogf("LSSUs received: %.10lu    %.10lu  \n", nLssuCount0, nLssuCount1);
        DemoLogf("MSUs  received: %.10lu    %.10lu  \n", nMsuCount0,  nMsuCount1);
        DemoLogf("Last FISU0 received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu0.bsn, fisu0.bib, fisu0.fsn, fisu0.fib, fisu0.li);
        DemoLogf("Last FISU1 received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu1.bsn, fisu1.bib, fisu1.fsn, fisu1.fib, fisu1.li);
      }

    }
    break;

  case 'L':
    {
      if (g_demoEnv.bUseSenders) {
        char strBuf[12];
        DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame0);
        OsaConsoleGetStr(strBuf);
        if (*strBuf) {
          g_cbHdlcFrame0 = (OTX_INT32)strtol(strBuf, NULL, 0);
          DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame0);
        } else {
          DemoLogf("HDLC Frame Length remains unchanged\n");
        }
      }
    }
    break;

  case 'S':
    {
      if (g_demoEnv.bUseSenders) {
        char strBuf[12];
        DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
        OsaConsoleGetStr(strBuf);
        if (*strBuf) {
          g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
          DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
        } else {
          DemoLogf("Sleep time remains unchanged\n");
        }
      }
    }
    break;

  case 'P':
    {
      if (g_demoEnv.bUseSenders) {
        OTX_RESULT result;
        OTX_INT32 i;
        OTX_INT32 busyCount = 0;
        clock_t startTime, stopTime;
        OtxHdlcSS7FisuS fisu1;
        OtxHdlcSS7FisuS fisu2;
        OTX_UINT8 sqBuf[10];
        OTX_UINT32 nDev = 0;
        
        g_BsnCheck = OTX_TRUE;
        g_BsnErrors = 0;
        
        for (i = 0; i < 10; i++) {
          sqBuf[i] = (OTX_UINT8)i;
        }
        
        fisu1.bib = 1;
        fisu1.fib = 1;
        fisu1.fsn = 0x7F;
        fisu1.bsn = 0x7F;
        
        memcpy(&fisu2, &fisu1, sizeof(OtxHdlcSS7FisuS));
        
        DemoLogf("Sending FISUs from dev%d\n", nDev);
        
        // Send MSU Frame(s)
        startTime = clock();
        for (i = 0; i < 1000; i++) {
          //result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, (OTX_UINT8 *)&fisu1, 3, &fisu2, DEMO_TASK_REF, OTX_TIME_INFINITY); 
          
          result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqBuf, (i == 999) ? 5 : 4, &fisu2, DEMO_TASK_REF, OTX_TIME_INFINITY); 
          if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
            // Driver IO Request buffer is full. We need to retransmit this frame.
            i--;
            busyCount++;
          } else {
            DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);
            
            fisu1.bsn++;
            memcpy(&fisu2, &fisu1, sizeof(OtxHdlcSS7FisuS));
            
            OsaThreadSleep(1);
          }
        }
        stopTime = clock();
        
        DemoLogf("[%d]Done g_BsnErrors=%d (BusyCount=%d, time=%2.1f seconds).\n", nDev, g_BsnErrors, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
      }      
    }
    
    break;

  case '+':
    {
      if (g_demoEnv.bUseSenders) {
        char strBuf[12];
        
        DemoLogf("Set the number of consecutive frame to sent when 'm', 't', or 'T' are pressed (%d)\n", g_coHdlcSenderFrames);
        OsaConsoleGetStr(strBuf);
        if (*strBuf) {
          g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
          DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
        } else {
          DemoLogf("Consecutive frame count remains unchanged\n");
        }
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

#if 0 // Only for debugging
  case 'd':
    DemoLogf("Dump DSP debug data\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 0);
    break;

  case 'b':
    DemoLogf("Init Bsp\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 1);
    break;
#endif


  case 't':
    {
      // Line 2 and 3 of thor-4-expresscard only supports receive no transmit
      if((g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) &&
         (nCurrentHdlcDev > 1)) {
        if(nCurrentHdlcDev < 4) {
          DemoLogf("Note: Li#%d Thor-4-ExpressCard is receive only\n", nCurrentHdlcDev);
        }
        else {
          DemoLogf("Invalid Li#%d Thor-4-ExpressCard has 4 lines\n", nCurrentHdlcDev);
          nCurrentHdlcDev = 0;
        }
      }

      if (g_demoEnv.bUseSenders) {
        
        if (g_demoEnv.bSs7FisuFiltering) {
          DemoLogf("In SS7 mode use the 'm' key to transmit an MSU\n");
        } else {
          OsaThreadStartEx(DemoHdlcTransmitFunction, (void *) &nCurrentHdlcDev);
        } 
      }
    }
    break;

  case 'C':
    {
      if (g_demoEnv.bUseSenders) {
//        OsaThreadStartEx(DemoHdlcTransmitFunction, (void *) &nCurrentHdlcDev);
        OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &nCurrentHdlcDev);
      }
    }
    break;


  case 'U':
    {
      if (g_demoEnv.bUseSenders) {
        // This performs a simulation of MSU and FISU sending. 
        // The Li is looped back to itself. 
        OTX_INT32 nLi = g_sqDspHdlcSender[nCurrentHdlcDev].nLi;
        g_nSs7SimulationFramesReceived[nCurrentHdlcDev] = 0;
        g_nSs7SimulationErrors[nCurrentHdlcDev] = 0;
        g_nPrevBsn[nCurrentHdlcDev] = 0x7F;
        
        
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        
        OsaThreadStartEx(DemoSs7SimulationFunction, (void *) &nCurrentHdlcDev);
      }
    }
    break;


  case 'u':
    {
      if (g_demoEnv.bUseSenders) {
        // This performs a simulation of MSU and FISU sending. 
        // The Li is looped back to itself. 
        OTX_INT32 nLi = g_sqDspHdlcSender[nCurrentHdlcDev].nLi;
        static OTX_BOOL g_bLooped[16];
        
        g_bLooped[nLi] = !g_bLooped[nLi];
        
        DemoLogf("Li#%d is now %s\n", g_bLooped[nLi] ? "Looped" : "Not looped");
        
        if (g_bLooped[nLi]) {
          result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        } else {
          result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOOPS_OFF); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        }
      }
    }
    break;


  case 'T':
    {
      // Line 2 and 3 of thor-4-expresscard only supports receive no transmit
      if((g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) &&
         (nCurrentHdlcDev > 1)) {
        if(nCurrentHdlcDev < 4) {
          DemoLogf("Note: Li#%d Thor-4-ExpressCard is receive only\n", nCurrentHdlcDev);
        }
        else {
          DemoLogf("Invalid Li#%d Thor-4-ExpressCard has 4 lines\n", nCurrentHdlcDev);
          nCurrentHdlcDev = 0;
        }
      }

      if (g_demoEnv.bUseSenders) {
        OTX_UINT8 *pBadMsg = (OTX_UINT8 *)"\x00\x01\x26\x3a\x4f\x02\x02\x68\x4d\x08\x03\x82\x90\x00";
        OTX_INT32 nMsgLen = 14;
        //OTX_UINT8 *pBadMsg = "\x04\x09\x7F";
        //OTX_INT32 nMsgLen = 3;
        OTX_HANDLE hHandle;
        
        hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
        
        DemoLogf("Sending test HDLC frame from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
        
        result = OtxHdlcSendData(hHandle, pBadMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
        
        DemoLogf("Done.\n");
      }
    }
    break;

  case 'F':
    {
      OTX_HANDLE hHandle;
      OTX_UINT32 filterMask;

      hHandle = g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle;

      result = OtxHdlcSS7GetFilter(hHandle, &filterMask); DemoCheckRc(hHandle, result);
      if (filterMask & OTX_HDLC_SS7_FILTER_FISU_ALL) {
        filterMask &= ~OTX_HDLC_SS7_FILTER_FISU_ALL;
      } else {
        filterMask |= OTX_HDLC_SS7_FILTER_FISU_ALL;
      }

      DemoLogf("Setting SS7 filter mask to %.4X for HDLC Receiver #%d (Dsp#%d, Li#%d)...", filterMask, nCurrentHdlcDev, g_sqDspHdlcReceiver[nCurrentHdlcDev].nDsp, g_sqDspHdlcReceiver[nCurrentHdlcDev].nLi);

      result = OtxHdlcSS7SetFilter(hHandle, filterMask); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'f':
    {
      if (g_demoEnv.bUseSenders) {
        OTX_HANDLE hHandle;
        
        hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
        DemoLogf("Start Sending FISUs (fsn=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
        
        if (nCurrentHdlcDev == 0) {
          g_tFisu0.fsn++;
          result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu0, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
        } else {
          g_tFisu1.fsn++;
          result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu1, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
        }
        // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
        
        DemoLogf("Done.\n");
      }
    }
    break;

  case 'i':
    {
      if (g_demoEnv.bUseSenders) {
        OTX_HANDLE hHandle;
        
        hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
        
        // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
        g_tLssu.fsn = g_tFisu0.fsn;
        g_tLssu.sf = g_tLssu.sf + 1; 
        g_tLssu.li = 1; // Short LSSU
        
        DemoLogf("Start Sending LSSUs (fsn=%d, sf=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", g_tLssu.fsn, g_tLssu.sf, nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
        
        result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tLssu, OTX_HDLC_SS7_CB_SHORT_LSSU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
        
        DemoLogf("Done.\n");
      }
    }
    break;

  case 'p':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle;


      DemoLogf("Reset HDLC Receiver #%d (Dsp#%d, Li#%d)...", g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxDrvReset(hHandle); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'm':
    {
      if (g_demoEnv.bUseSenders) {
        if (nCurrentHdlcDev == 0) {
          g_tFisu0.fsn++;
        } else {
          g_tFisu1.fsn++;
        }
        OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &nCurrentHdlcDev);
      }
    }
    break;

  case 'r':
    {
      if (g_demoEnv.bUseSenders) {
        
        OTX_UINT8 *pRepeatFrame = (OTX_UINT8 *)"\x10\x11\x12\x13\x14";
        OTX_INT32 nMsgLen = 5;
        OTX_HANDLE hHandle;
        
        hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
        
        DemoLogf("Start Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
        
        result = OtxHdlcStartRepeatData(hHandle, pRepeatFrame, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
        
        DemoLogf("Done.\n");
      }
    }
    break;
  case 'R':
    {
      if (g_demoEnv.bUseSenders) {
        OTX_HANDLE hHandle;
        
        hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
        
        DemoLogf("Stop Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
        
        result = OtxHdlcStopRepeatData(hHandle); DemoCheckRc(hHandle, result);
        
        DemoLogf("Done.\n");
      }
    }
    break;


  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      printf("Set Clock Source:\n");
      printf("0 = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");
      printf("1 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to FA0 receive span\n");
      printf("2 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to FA1 receive span\n");

      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      case 1:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
        break;
      case 2:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
        break;
      default:
        printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      }

      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_H100, eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      // Setting the clock source also performs a reset of the E1T1 devices. Therefore we need to reconfigure the E1T1 devices here.
      DemoConfigureE1T1Devices();

      printf("Done.\n");
    }
    break;
  
  case '$':
    {
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source ...\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("RX Line frequency is %d Hz (nFrequencyCount=%d)\n", nFrequency, nFrequencyCount);
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'n':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev++;
      if (nCurrentHdlcDev > (g_demoEnv.coLi*g_demoEnv.coHdlcPerLi)-1) {
        // Wrap 
        nCurrentHdlcDev = 0;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev--;
      if (nCurrentHdlcDev < 0) {
        // Wrap
        nCurrentHdlcDev = (g_demoEnv.coLi*g_demoEnv.coHdlcPerLi)-1;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
    }
    break;

#if 0 // Just for debugging
  case '#':
    {
      DemoLogf("Performing an illegal operation. Crash test\n");
      memcpy(NULL, &nCurrentHdlcDev, 200);
    }
    break;
#endif
  case 'a':
    {
      // Specify whether the HDLC Sender device should send us the 
      // OTX_HDLC_SENDER_EC_FRAME_SENT event
      OTX_BOOL   bHdlcFrameSentEvents = OTX_TRUE;
      OTX_HANDLE hHandle;
      OTX_ATTRIB_ID nAttribute;

      // Start 16 bit frame count rollover variable with max value, when
      // OTX_HDLC_SENDER_EC_FRAME_COUNT_ZERO is processed the count 
      // will be increamented by one which will synchronize the application
      // with the DSP device frame counter.
      g_coHdlcFramesSentRollOverCount[nCurrentHdlcDev] = (0-1);
      DemoLogf("HDLC Sender[%d] Set rollover count to %x\n", 
               nCurrentHdlcDev,
               g_coHdlcFramesSentRollOverCount[nCurrentHdlcDev]);

      g_coHdlcFramesSentRollOverCount[nCurrentHdlcDev] = (0-1);
      DemoLogf("OTX_HDLC_SENDER_EC_FRAME_COUNT_ZERO will synchronize the count.\n");


      nAttribute = OTX_ATTR_HDLC_SENDER_GENERATE_FRAME_SENT_EVENT;
      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      result = OtxDrvGetAttributeValueBool(hHandle, 
                                           nAttribute, 
                                           &bHdlcFrameSentEvents); 
      
      // Toggle the setting
      bHdlcFrameSentEvents = !bHdlcFrameSentEvents;

      DemoLogf("HDLC Sender[%d] Setting frame sent event %s\n", 
               nCurrentHdlcDev, 
               (bHdlcFrameSentEvents) ? "TRUE": "FALSE");

      result = OtxDrvSetAttributeValueBool(hHandle, 
                                           nAttribute, 
                                           bHdlcFrameSentEvents); 
      DemoCheckRc(hHandle, result);
    }
    break;

  case 'I':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'O':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;


    
        
  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      printf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        printf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoLogf("T2p5x4Hdlc, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("HDLC Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (((g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) && (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXHDLC)) ||
      ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) && (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXHDLC))) {
    // The user did not supply a program module for the ExpressCard, so set the default to a proper module
    g_demoEnv.eProgramModule = OTX_DSP_SPM_OTXHDLC_INTMEM;
  }

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  if (g_demoEnv.bSuperChannel) {
    DemoLogf("1 superchannel per DSP (%d x 64 kbps capacity)\n", g_demoEnv.nTimeSlots);
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  {
    char *pszClockSource;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      pszClockSource = "Internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
      pszClockSource = "Line Interface 0";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_1:
      pszClockSource = "Line Interface 1";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_2:
      pszClockSource = "External";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_3:
      pszClockSource = "Split Clock, both Line Interface are synchronized to its T1E1 span";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_4:
      pszClockSource = "Split Clock, Li0 synchronized to its T1E1 span. Li1 internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_5:
      pszClockSource = "Split Clock, Li1 synchronized to its T1E1 span. Li0 internal";
      break;
    default:
      pszClockSource = "Unknown";
      break;
    }

    DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
    DemoLogf("Clock Source: %s\n", pszClockSource);
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  {
    char szClockSource[64];
    OtxBrdClockModeE     clkMode = OTX_CLOCK_MODE_H100_ETCF;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
      sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);

    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
      clkMode = OTX_CLOCK_MODE_NA; // This board type does not have the H.100 bus
    } else {
      clkMode = OTX_CLOCK_MODE_H100_ETCF; 
    }
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, clkMode, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
    // On these two boards types we need to send this signal to the DSPs after setting the clock
    if (g_hPhysDevs.hVidarAsmController0) {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
  }

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszCpldBuildNumbers;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }
  }
  
  g_tFisu1.bib = 1;
  g_tFisu1.fib = 1;
  g_tFisu1.fsn = 0x7F;
  g_tFisu1.bsn = 0x7F;

  g_tFisu0.bib = 1;
  g_tFisu0.fib = 1;
  g_tFisu0.fsn = 0x7F;
  g_tFisu0.bsn = 0x7F;

  DemoLogf("\nPress 'h' for help\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

#if 1
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }
#else // Just for testing. Exiting automatically after 300ms
  OsaThreadSleep(300);
#endif

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
